################################################################################
################ SUPPLEMENTARY SCRIPT FOR APHID MANUSCRIPT #####################
################################################################################


## Written by Shaun Wilkinson 2019-02-24
## Contact shaunpwilkinson AT gmail DOT com
## Contains scripts for plotting HMMs & profile HMMS
## Also shows benchmarking working and results using HMM R package and HMMER
## All working done on Lenovo T530 laptop with intel i7-3720QM 2.6 GHz
## OS: Ubuntu 16.04 LTS

## Install required packages:
if(!("devtools" %in% list.files(.libPaths()))) install.packages("devtools")
if(!("aphid" %in% list.files(.libPaths()))) devtools::install_github("shaunpwilkinson/aphid")
if(!("insect" %in% list.files(.libPaths()))) devtools::install_github("shaunpwilkinson/insect")
if(!("AlignStat" %in% list.files(.libPaths()))) devtools::install_github("TS404/AlignStat")
if(!("ape" %in% list.files(.libPaths()))) install.packages("ape")
if(!("HMM" %in% list.files(.libPaths()))) install.packages("HMM")


#################### 1. CREATE PLOTS IN MAIN DOCUMENT ##########################

## Standard HMM (splice site example)
library("aphid")
## Define symbols and hidden states
states <- c("Begin", "EXON", "SPLICE SITE", "INTRON")
residues <- c("A", "C", "G", "T")
## Define transition probability matrix
A <- matrix(c(0, 0, 0, 0.1, 1, 0.9, 0, 0, 0, 0.1, 0, 0, 0, 0, 1, 0.9), nrow = 4)
dimnames(A) <- list(from = states, to = states)
## Define emission probability matrix
E <- matrix(c(rep(0.25, 4), 0.05, 0, 0.95, 0, 0.4, 0.1, 0.1, 0.4), nrow = 3, byrow = TRUE)
dimnames(E) <- list(states = states[-1], residues = residues)
## Create the HMM object
x <- structure(list(A = A, E = E), class = "HMM")
## Read in the example sequence from Eddy (2004)
s <- strsplit("CTTCATGTGAAAGCAGACGTAAGTCA", split = "")[[1]]
## Print most likely path of hidden states
## Note: add 1 to Viterbi path to make consistent with R's indexing style
paste0(c("E", "S", "I")[aphid::Viterbi(x, s)$path + 1], collapse = "")
pos <- aphid::posterior(x, s)
spliceprobs <- pos[2, ]
names(spliceprobs) <- s
#plot(seq_along(pos[2, ]), pos[2, ], type = "h", lwd = 10)
setEPS()
postscript("spliceprobs.eps", width = 7, height = 3.2, family = "Times")
barplot(spliceprobs, ylim = c(0, 0.5), xlab = "", xaxt = "n", yaxt = "n", col = "black")
dev.off()

## Add nominal values to show all transitions (show faint lines for zero probabilities)
x$A <- x$A + 0.01
## Normalize so that probabilities sum to 1
x$A <- x$A/apply(x$A, 1, sum)

## Plot the model
setEPS()
postscript("Fig1a.eps", width = 7, 4, family = "Times")
plot(x, textexp = 1, begin = TRUE, arrexp = 1.2)
### Annotate non-zero transition probabilities
text(x = 0.24, y = 0.6, labels = "1.0")
text(x = 0.495, y = 0.6, labels = "0.1")
text(x = 0.745, y = 0.6, labels = "1.0")
text(x = 0.24, y = 0.495, labels = "0.9")
text(x = 0.745, y = 0.495, labels = "0.9")
text(x = 0.501, y = 0.325, labels = "0.1")##
dev.off()


## Profile HMM (globin example from Durbin et al. 1998)
data(globins)
globins.PHMM <- aphid::derivePHMM(globins, residues = "AMINO",
                                  pseudocounts = "Laplace",
                                  DI = TRUE, ID = TRUE)
setEPS()
postscript("Fig1b.eps", width = 7, 3, family = "Times")
plot(globins.PHMM)
dev.off()
## Note: plots were marked up with Inkscape for publication

######################## 2. STANDARD HMM BENCHMARKING ##########################

## Compare aphid and HMM packages for speed and accuracy
## Note HMM package only accepts single sequences for Baum Welch/Viterbi training
## so simulate one long sequence for training
## dishonest casino example from Durbin et al. (1998)

library("aphid")
## Define symbols and hidden states
states <- c("Begin", "Fair", "Loaded")
residues <- paste(1:6)
## Define transition probability matrix
A <- matrix(c(0, 0, 0, 0.99, 0.95, 0.1, 0.01, 0.05, 0.9), nrow = 3)
dimnames(A) <- list(from = states, to = states)
## Define emission probability matrix
E <- matrix(c(rep(1/6, 6), rep(1/10, 5), 1/2), nrow = 2, byrow = TRUE)
dimnames(E) <- list(states = states[-1], residues = residues)
## Create the HMM object
casino.HMM <- structure(list(A = A, E = E), class = "HMM")
## Plot the model
plot(casino.HMM, textexp = 1.5)
## Annotate with non-zero transition probabilities
text(x = 0.02, y = 0.5, labels = "0.95")
text(x = 0.51, y = 0.5, labels = "0.90")
text(x = 0.5, y = 0.9, labels = "0.05")
text(x = 0.5, y = 0.1, labels = "0.10")
## Simulate a sequence of one million rolls
set.seed(999)
simseq <- aphid::generate(casino.HMM, size = 1000000)
## Time aphid to find optimal path on single processor
system.time(optpath <- aphid::Viterbi(casino.HMM, simseq)$path)
## 0.739s
## Note that aphid can also give the log-odds score by running:
Viterbi(casino.HMM, simseq)$score
## Convert path from integer vector to character strings for comparison
## Note: add 1 to Viterbi path to make consistent with R's indexing style
optpath <- c("Fair", "Loaded")[optpath + 1L]
## Time aphid to find full probability of sequence given model
system.time(fullprob <- aphid::forward(casino.HMM, simseq))
## 0.429s (fullprob$score = -1740601)

## Initiallise the same model with the HMM package
A <- A[-1, ]
A <- A[, -1]
casino.HMM2 <- HMM::initHMM(States = c("Fair", "Loaded"),
                            Symbols = paste(1:6),
                            startProbs = c(0.99, 0.01),
                            transProbs = A, emissionProbs = E)
## Time HMM package to find optimal path
system.time(optpath2 <- HMM::viterbi(casino.HMM2, simseq))
## 25.98s
all(optpath == optpath2) # TRUE - gives the same result
## Time HMM package to compute full probability
system.time(fullprob2 <- HMM::forward(casino.HMM2, simseq))
# 15.797s
logsum(fullprob2[, 1000000]) # -1740601 - same as aphid
## Note: aphid is faster due to implementation of intensive recursions in C++


######################### 3. PROFILE HMM BENCHMARKING ##########################

## Compare aphid with HMMER on globin seed alignment from PFAM (PF00042)
## Downloaded from https://pfam.xfam.org/family/PF00042#tabview=tab3
##   on 2019-02-24 in FASTA format with "-" gaps
library(aphid)
alignment <- insect::readFASTA("PF00042_seed.txt", residues = "AA", alignment = TRUE)
## Time aphid to derive profile HMM
system.time(aphidphmm <- derivePHMM(alignment, inserts = "threshold",
                                    threshold = 0.5, seqweights = NULL,
                                    pseudocounts = "Laplace"))
## 0.020 s
## Build system call and time HMMER to derive the same PHMM
sc <- "hmmbuild --amino --wnone --enone --plaplace --cpu 1 --fragthresh 0 PF00042.hmm PF00042_seed.txt"
system.time(system(sc)) # 0.026 s
## Compare transition and emission probabilities
hmmerphmm <- readPHMM("PF00042.hmm")
aphidphmm$A[, 1:5]
hmmerphmm$A[, 1:5]
aphidphmm$E[, 1:5]
hmmerphmm$E[, 1:5]
## Models are essentially the same (some rounding diffs)
## Deconstruct alignment to sequence list and write out to fasta
seqlist <- aphid::unalign(alignment)
insect::writeFASTA(seqlist, file = "PF00042.fa")
## Time hmmer to find full probability of each sequence given model
system.time(system("hmmsearch --max --nonull2 --cpu 1 PF00042.hmm PF00042.fa > PF00042.txt"))
## 0.072 s on a single processor
## Read in results and extract bit-scores
output <- scan("PF00042.txt", what = "", sep = "\n")
head(output, 100)
outnames <- sub(".+    1.0  1  ([^ ]+) +", "\\1", output[17:89])
outscores <- as.numeric(sub(" +[^ ]+ +([^ ]+) +.+", "\\1", output[17:89]))
## Time aphid to complete the same task
score <- function(x, m) aphid::forward(m, x, odds = TRUE)$score
system.time(aphscores <- sapply(seqlist, score, aphidphmm))
## 0.613 s on a single processor
## HMMER is faster (applies optimizaion heuristics, SIMD processing)
## Reorder aphid scores to match HMMER output
aphscores <- aphscores[outnames]
plot(outscores, aphscores, xlab = "HMMER bit-scores", ylab = "aphid bit-scores")
## Gives similar log odds scores, differences due to rounding,
## differenct heuristics and methods of calculating background emission
## and transition probabilities, etc.

## Align sequences using model as a guide
## Note: HMMER cannot currently do global alignments (see HMMER docs)
system.time(realignment1 <- aphid::align(seqlist, model = aphidphmm, cores = 1))
## 0.72 s, aphid global alignment has 131 columns
sc <- "hmmalign -o PF00042.afa --outformat afa --amino PF00042.hmm PF00042.fa"
system.time(system(sc))
## 0.044 s, faster due to local alignment
realignment2 <- insect::readFASTA("PF00042.afa", residues = "AA", alignment = TRUE)
## HMMER alignment has 151 columns
## original has 124 columns
## Compare each alignment with original using sum-of-pairs score
insect::writeFASTA(realignment1, file = "realignment1.afa")
insect::writeFASTA(realignment2, file = "realignment2.afa")
AlignStat::compare_alignments("PF00042_seed.txt", "realignment1.afa")$similarity_score
## 0.980
AlignStat::compare_alignments("PF00042_seed.txt", "realignment2.afa")$similarity_score
## 0.958
## aphid alignment is closer to the original reference

## Final note: HMMER has been extensively optimized for computationally-intensive homology
## searches, and aphid is not designed to compete with HMMER on speed and efficiency.
## Rather, aphid is a complementary tool that offers fine-scale control of HMMs and
## profile HMMs (eg model training, plotting, IO, etc) in a flexible environment.
## All feedback welcome. Please direct bug reports to
## https://github.com/shaunpwilkinson/aphid/issues


############################ 4. REFERENCES #####################################

##   Durbin R, Eddy SR, Krogh A, Mitchison G (1998) Biological
##   sequence analysis: probabilistic models of proteins and nucleic acids.
##   Cambridge University Press, Cambridge, United Kingdom.
##
##   Eddy S (2004) What is a hidden Markov model? Nature Biotechnology, 22, 1315–1316.
##
##   Eddy S (2009) A new generation of homology search tools based on probabilistic
##   inference. Genome Informatics, 23, 205–211.
##
##   Himmelmann L (2010) HMM: Hidden Markov Models. R package version 1.0.
##   https://CRAN.R-project.org/package=HMM.
##
##   Paradis E, Claude J, Strimmer K, (2004) APE: analyses of phylogenetics
##   and evolution in R language. Bioinformatics, 20, 289-290.
##
##   Paradis E (2012) Analysis of Phylogenetics and Evolution with R
##   (Second Edition). Springer, New York.
##
##   Shafee T, Cooke I (2016) AlignStat: a web-tool and R package for
##   statistical comparison of alternative multiple sequence alignments.
##   BMC Bioinformatics, 17, 434.

################################################################################
